{-# LANGUAGE FunctionalDependencies #-}
module Types.FunctionIO
  ( IsFunctionIO(..)
  , ListDirectory(..)
  , IsDirectory(..)
  , ReadTextFile(..)
  , OpenImage(..)
  , SaveImage(..)
  , SaveTextFile(..)
  , CopyFile(..)
  , MakeDir(..)
  ) where

import Prelude hiding (String, FilePath)

import Types.Values
import Types.Value (Valuable)

import Data.Text (Text)
import qualified Data.Text.IO as T
import qualified Codec.Picture as CP
import System.Directory (listDirectory, doesDirectoryExist, createDirectory, copyFile)
import Language.Haskell.TH.Syntax (Lift)

class (Show f, Lift f, Valuable a, Valuable b) => IsFunctionIO f a b | f -> a b where
  evalFunctionIO :: f -> a -> IO b
  functionIOTouchesFilesystem :: f -> Bool

data ListDirectory = ListDirectory deriving (Show, Lift)
instance IsFunctionIO ListDirectory FilePath [FilePath] where
  evalFunctionIO ListDirectory (StringWrapper s) =
    map StringWrapper <$> listDirectory s
  functionIOTouchesFilesystem ListDirectory = False

data IsDirectory = IsDirectory deriving (Show, Lift)
instance IsFunctionIO IsDirectory FilePath Bool where
  evalFunctionIO IsDirectory (StringWrapper s) =
    doesDirectoryExist s
  functionIOTouchesFilesystem IsDirectory = False

data ReadTextFile = ReadTextFile deriving (Show, Lift)
instance IsFunctionIO ReadTextFile FilePath Text where
  evalFunctionIO ReadTextFile (StringWrapper s) =
    T.readFile s
  functionIOTouchesFilesystem ReadTextFile = False

data OpenImage = OpenImage deriving (Show, Lift)
instance IsFunctionIO OpenImage FilePath Image where
  evalFunctionIO OpenImage (StringWrapper s) = do
    imageOrig <- CP.readImage s
    case imageOrig of
      Left e -> error ("unexpected error: " ++ e)
      Right image -> pure $ ImageWrapper $ CP.convertRGB8 image
  functionIOTouchesFilesystem OpenImage = False

data SaveImage = SaveImage deriving (Show, Lift)
instance IsFunctionIO SaveImage (Image, FilePath) () where
  evalFunctionIO SaveImage (ImageWrapper image, StringWrapper s) =
    CP.saveJpgImage 90 s $ CP.ImageRGB8 image
  functionIOTouchesFilesystem SaveImage = True

data SaveTextFile = SaveTextFile deriving (Show, Lift)
instance IsFunctionIO SaveTextFile (Text, FilePath) () where
  evalFunctionIO SaveTextFile (t, StringWrapper s) =
    T.writeFile s t
  functionIOTouchesFilesystem SaveTextFile = True

data CopyFile = CopyFile deriving (Show, Lift)
instance IsFunctionIO CopyFile (FilePath, FilePath) () where
  evalFunctionIO CopyFile (StringWrapper source, StringWrapper target) =
    copyFile source target
  functionIOTouchesFilesystem CopyFile = True

data MakeDir = MakeDir deriving (Show, Lift)
instance IsFunctionIO MakeDir FilePath () where
  evalFunctionIO MakeDir (StringWrapper s) =
    createDirectory s
  functionIOTouchesFilesystem MakeDir = True
